@@ -255,14 +255,15 @@ void parse_options_set_display_postscript(Parse_Options opts, int dummy) {
 	opts->display_postscript = dummy;
 }
 
-int parse_options_get_display_postscript(Parse_Options opts) {
+int parse_options_get_display_postscript(Parse_Options opts)
+{
 	return opts->display_postscript;
 }
 
 void parse_options_set_display_constituents(Parse_Options opts, int dummy)
 {
 	if ((dummy < 0) || (dummy > 3)) {
-	   prt_error("Possible values for constituents: \n"
+		prt_error("Possible values for constituents: \n"
 	             "   0 (no display)\n"
 	             "   1 (treebank style, multi-line indented)\n"
 	             "   2 (flat tree, square brackets)\n"
@@ -272,7 +273,8 @@ void parse_options_set_display_constituents(Parse_Options opts, int dummy)
 	else opts->display_constituents = dummy;
 }
 
-int parse_options_get_display_constituents(Parse_Options opts) {
+int parse_options_get_display_constituents(Parse_Options opts)
+{
 	return opts->display_constituents;
 }
 
@@ -530,8 +532,6 @@ Sentence sentence_create(char *input_string, Dictionary dict)
 		return NULL;
 	}
 
-	error_report_set_sentence(sent);
-
 	sent->q_pruned_rules = FALSE; /* for post processing */
 	sent->is_conjunction = (char *) xalloc(sizeof(char)*sent->length);
 	set_is_conjunction(sent);
@@ -604,7 +604,6 @@ void sentence_delete(Sentence sent)
 	free_analyze(sent);
 	xfree(sent->is_conjunction, sizeof(char)*sent->length);
 	xfree((char *) sent, sizeof(struct Sentence_s));
-	error_report_set_sentence(NULL);
 }
 
 int sentence_length(Sentence sent)
@@ -696,9 +695,13 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 		/* We know that sent->num_linkages_found is bogus, possibly negative */
 		sent->num_linkages_found = opts->linkage_limit;
 		if (opts->verbosity > 1)
-		  prt_error("Warning: Count overflow.\n"
+		{
+			err_ctxt ec;
+			ec.sent = sent;
+			err_msg(&ec, Warn, "Warning: Count overflow.\n"
 			  "Considering a random subset of %d of an unknown and large number of linkages\n",
 				opts->linkage_limit);
+		}
 	}
 	N_linkages_found = sent->num_linkages_found;
 
@@ -713,8 +716,11 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 	if (N_linkages_found > opts->linkage_limit)
 	{
 		N_linkages_alloced = opts->linkage_limit;
-		if (opts->verbosity > 1) {
-		  prt_error("Warning: Considering a random subset of %d of %d linkages\n",
+		if (opts->verbosity > 1)
+		{
+			err_ctxt ec;
+			ec.sent = sent;
+			err_msg(&ec, Warn, "Warning: Considering a random subset of %d of %d linkages\n",
 				  N_linkages_alloced, N_linkages_found);
 		}
 	}
@@ -821,7 +827,9 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 			 * of Nottingham and you be Robin Hood a little while and kill
 			 * me.
 			 */
-			prt_error("Error: None of the linkages is canonical\n"
+			err_ctxt ec;
+			ec.sent = sent;
+			err_msg(&ec, Error, "Error: None of the linkages is canonical\n"
 			          "\tN_linkages_post_processed=%d "
 			          "N_linkages_found=%d\n",
 			          N_linkages_post_processed,
@@ -829,8 +837,11 @@ static void post_process_linkages(Sentence sent, Parse_Options opts)
 		}
 	}
 
-	if (opts->verbosity > 1) {
-		prt_error("Info: %d of %d linkages with no P.P. violations\n",
+	if (opts->verbosity > 1)
+	{
+		err_ctxt ec;
+		ec.sent = sent;
+		err_msg(&ec, Info, "Info: %d of %d linkages with no P.P. violations\n",
 				N_valid_linkages, N_linkages_post_processed);
 	}
 
