@@ -14,6 +14,7 @@
 #endif
 
 #include "api.h"
+#include "error.h"
 #include "jni-client.h"
 #include "utilities.h"
 
@@ -50,10 +51,17 @@ static per_thread_data * global_ptd = NULL;
 static per_thread_data * get_ptd(JNIEnv *env, jclass cls)
 {
 #ifdef USE_PTHREADS
-	return pthread_getspecific(java_key);
+	per_thread_data *ptd = pthread_getspecific(java_key);
 #else
-	return global_ptd;
+	per_thread_data *ptd = global_ptd;
 #endif
+	if (!ptd) Java_org_linkgrammar_LinkGrammar_init(env, cls);
+#ifdef USE_PTHREADS
+	ptd = pthread_getspecific(java_key);
+#else
+	ptd = global_ptd;
+#endif
+	return ptd;
 }
 
 static void setup_panic_parse_options(Parse_Options opts)
@@ -69,7 +77,7 @@ static void setup_panic_parse_options(Parse_Options opts)
 	parse_options_set_verbosity(opts,0);
 }
 
-static void test(void)
+static inline void test(void)
 {
 #ifdef DEBUG
 	printf("%d\n",word_contains("said",PAST_TENSE_FORM_MARKER,dict));
@@ -129,6 +137,42 @@ static per_thread_data * init(JNIEnv *env, jclass cls)
 	return ptd;
 }
 
+static void finish(per_thread_data *ptd)
+{
+	if (ptd->sent)
+		sentence_delete(ptd->sent);
+	ptd->sent = NULL;
+
+#if DO_PHRASE_TREE
+	if (tree)
+		linkage_free_constituent_tree(tree);
+	tree = NULL;
+#endif
+
+	if (ptd->linkage)
+		linkage_delete(ptd->linkage);
+	ptd->linkage = NULL;
+
+	dictionary_delete(ptd->dict);
+	ptd->dict = NULL;
+
+	parse_options_delete(ptd->opts);
+	ptd->opts = NULL;
+
+	parse_options_delete(ptd->panic_parse_opts);
+	ptd->panic_parse_opts = NULL;
+
+#ifdef USE_PTHREADS
+	pthread_setspecific(java_key, NULL);
+#else
+	global_ptd = NULL;
+#endif
+	free(ptd);
+}
+
+/* ================================================================= */
+/* Misc utilities */
+
 #ifdef DEBUG_DO_PHRASE_TREE
 static void r_printTree(CNode* cn, int level)
 {
@@ -224,7 +268,7 @@ static void jParse(JNIEnv *env, per_thread_data *ptd, char* inputString)
 			parse_options_resources_exhausted(opts) &&
 			parse_options_get_panic_mode(opts))
 	{
-		print_total_time(opts);
+		parse_options_print_total_time(opts);
 		if (jverbosity > 0) prt_error("Warning: Entering \"panic\" mode...\n");
 		parse_options_reset_resources(ptd->panic_parse_opts);
 		parse_options_set_verbosity(ptd->panic_parse_opts, jverbosity);
@@ -256,33 +300,7 @@ static void makeLinkage(per_thread_data *ptd)
 	}
 }
 
-static void finish(per_thread_data *ptd)
-{
-	if (ptd->sent)
-		sentence_delete(ptd->sent);
-	ptd->sent = NULL;
-
-#if DO_PHRASE_TREE
-	if (tree)
-		linkage_free_constituent_tree(tree);
-	tree = NULL;
-#endif
-
-	if (ptd->linkage)
-		linkage_delete(ptd->linkage);
-	ptd->linkage = NULL;
-
-	dictionary_delete(ptd->dict);
-	ptd->dict = NULL;
-
-	parse_options_delete(ptd->opts);
-	ptd->opts = NULL;
-
-	parse_options_delete(ptd->panic_parse_opts);
-	ptd->panic_parse_opts = NULL;
-}
-
-/* =========================================================================== */
+/* ================================================================ */
 /* Java JNI wrappers */
 
 /*
@@ -293,7 +311,7 @@ static void finish(per_thread_data *ptd)
 JNIEXPORT jstring JNICALL
 Java_org_linkgrammar_LinkGrammar_getVersion(JNIEnv *env, jclass cls)
 {
-	const char *s = "link-grammar-" LINK_VERSION_STRING;
+	const char *s = linkgrammar_get_version();
 	jstring j = (*env)->NewStringUTF(env, s);
 	return j;
 }
@@ -321,7 +339,7 @@ Java_org_linkgrammar_LinkGrammar_setDictionariesPath(JNIEnv *env,
 	// Java passes null pointers as the string "null"
 	if (nativePath && strcmp(nativePath, "null")) 
 	{
-		set_data_dir(nativePath);
+		dictionary_set_data_dir(nativePath);
 	}
 	(*env)->ReleaseStringUTFChars(env,path, nativePath);
 }
@@ -337,9 +355,12 @@ Java_org_linkgrammar_LinkGrammar_init(JNIEnv *env, jclass cls)
 #ifdef USE_PTHREADS
 	per_thread_data *ptd;
 	pthread_once(&java_key_once, java_key_alloc);
+	ptd = pthread_getspecific(java_key);
+	if (ptd) return;
 	ptd = init(env, cls);
 	pthread_setspecific(java_key, ptd);
 #else
+	if (global_ptd) return;
 	global_ptd = init(env, cls);
 #endif
 }
@@ -627,7 +648,7 @@ Java_org_linkgrammar_LinkGrammar_isPastTenseForm(JNIEnv *env, jclass cls, jstrin
 
 	per_thread_data *ptd = get_ptd(env, cls);
 	const char *cStr = (*env)->GetStringUTFChars(env,str,0);
-	if (is_past_tense_form(cStr,ptd->dict) == 1)
+	if (dictionary_is_past_tense_form(ptd->dict, cStr) == 1)
 		rv = TRUE;
 	(*env)->ReleaseStringUTFChars(env,str,cStr);
 	return rv;
@@ -645,7 +666,7 @@ Java_org_linkgrammar_LinkGrammar_isEntity(JNIEnv *env, jclass cls, jstring str)
 
 	per_thread_data *ptd = get_ptd(env, cls);
 	const char *cStr = (*env)->GetStringUTFChars(env,str,0);
-	if (is_entity(cStr,ptd->dict) == 1)
+	if (dictionary_is_entity(ptd->dict, cStr) == 1)
 		rv = TRUE;
 	(*env)->ReleaseStringUTFChars(env,str,cStr);
 	return rv;
