@@ -13,12 +13,8 @@
 
 #include <link-grammar/api.h>
 
-typedef struct connector_table_s connector_table;
-struct connector_table_s
-{
-	int s_table_size;
-	Connector ** table;
-};
+#define CONTABSZ 8192
+typedef Connector * connector_table;
 
 typedef struct disjunct_dup_table_s disjunct_dup_table;
 struct disjunct_dup_table_s
@@ -123,6 +119,32 @@ struct prune_context_s
 
 */
 
+/**
+ * This hash function only looks at the leading upper case letters of
+ * the connector string, and the label fields.  This ensures that if two
+ * strings match (formally), then they must hash to the same place.
+ */
+static inline int hash_S(Connector * c)
+{
+	const char *s;
+	unsigned int i;
+
+	if (-1 != c->hash) return c->hash;
+
+	/* djb2 hash */
+	i = 5381;
+	i = ((i << 5) + i) + c->label;
+	s = c->string;
+	while(isupper((int)*s)) /* connector tables cannot contain UTF8, yet */
+	{
+		i = ((i << 5) + i) + *s;
+		s++;
+	}
+	c->hash = (i & (CONTABSZ-1));
+	c->prune_string = s;
+	return c->hash;
+}
+
 /**
  * This is almost identical to match().  Its reason for existance
  * is the rather subtle fact that with "and" can transform a "Ss"
@@ -137,13 +159,15 @@ int prune_match(int dist, Connector *a, Connector *b)
 	int x, y;
 
 	if (a->label != b->label) return FALSE;
-	x = a->priority;
-	y = b->priority;
+
+	hash_S(a);
+	hash_S(b);
+	if (a->hash != b->hash) return FALSE;
 
 	s = a->string;
 	t = b->string;
 
-	while(isupper((int)*s) || isupper((int)*t)) /* connector names are not yet UTF8-capable */
+	while(s < a->prune_string || t < b->prune_string)
 	{
 		if (*s != *t) return FALSE;
 		s++;
@@ -154,8 +178,11 @@ int prune_match(int dist, Connector *a, Connector *b)
 		   s, t, x, y, a->length_limit, b->length_limit, dist); */
 	if (dist > a->length_limit || dist > b->length_limit) return FALSE;
 
+	x = a->priority;
+	y = b->priority;
 
-	if ((x==THIN_priority) && (y==THIN_priority)) {
+	if ((x == THIN_priority) && (y == THIN_priority))
+	{
 #if defined(PLURALIZATION)
 /*
 		if ((*(a->string)=='S') && ((*s=='s') || (*s=='p')) &&  (*t=='p')) {
@@ -182,266 +209,224 @@ int prune_match(int dist, Connector *a, Connector *b)
    The third line above ensures that the connector is either "S" or "SI".
 */
 #endif
-		while ((*s!='\0') && (*t!='\0')) {
+		while ((*s != '\0') && (*t != '\0'))
+		{
 			if ((*s == '*') || (*t == '*') ||
-				((*s == *t) && (*s != '^'))) {
+				((*s == *t) && (*s != '^')))
+			{
 			  /* this last case here is rather obscure.  It prevents
 				 '^' from matching '^'.....Is this necessary?
 					 ......yes, I think it is.   */
 				s++;
 				t++;
-			} else return FALSE;
+			}
+			else
+				return FALSE;
 		}
 		return TRUE;
-	} else if ((x==UP_priority) && (y==DOWN_priority)) {
-		while ((*s!='\0') && (*t!='\0')) {
-			if ((*s == *t) || (*s == '*') || (*t == '^')) {
+	}
+	else if ((x == UP_priority) && (y == DOWN_priority))
+	{
+		while ((*s!='\0') && (*t!='\0'))
+		{
+			if ((*s == *t) || (*s == '*') || (*t == '^'))
+			{
 				/* that '^' should match on the DOWN_priority
 				   node is subtle, but correct */
 				s++;
 				t++;
-			} else return FALSE;
+			}
+			else
+				return FALSE;
 		}
 		return TRUE;
-	} else if ((y==UP_priority) && (x==DOWN_priority)) {
-		while ((*s!='\0') && (*t!='\0')) {
-			if ((*s == *t) || (*t == '*') || (*s == '^')) {
+	}
+	else if ((y == UP_priority) && (x == DOWN_priority))
+	{
+		while ((*s!='\0') && (*t!='\0'))
+		{
+			if ((*s == *t) || (*t == '*') || (*s == '^'))
+			{
 				s++;
 				t++;
-			} else return FALSE;
+			}
+			else
+				return FALSE;
 		}
 		return TRUE;
-	} else return FALSE;
-}
-
-/** 
- * This function removes all connectors from the set S
- */
-static void free_S(connector_table *ct)
-{
-	int i;
-	for (i = 0; i < ct->s_table_size; i++)
-	{
-		if (ct->table[i] == NULL) continue; /* a prehaps stupid optimization */
-		free_connectors(ct->table[i]);
-		ct->table[i] = NULL;
 	}
+	else
+		return FALSE;
 }
 
-static void connector_table_delete(connector_table *ct)
+static inline void zero_connector_table(connector_table *ct)
 {
-	free_S(ct);
-	xfree((char *)ct->table, ct->s_table_size * sizeof (Connector *));
-	free(ct);
+	memset(ct, 0, sizeof(Connector *) * CONTABSZ);
 }
 
-/**
- * This hash function only looks at the leading upper case letters of
- * the connector string, and the label fields.  This ensures that if two
- * strings match (formally), then they must hash to the same place.
- */
-static int hash_S(connector_table *ct, Connector * c)
-{
-	const char *s;
-	int i;
-	i = c->label;
-	s = c->string;
-	while(isupper((int)*s)) /* connector tables cannt contain UTF8, yet */
-	{
-		i = i + (i<<1) + randtable[(*s + i) & (RTSIZE-1)];
-		s++;
-	}
-	return (i & (ct->s_table_size-1));
-}
-
-/** 
- * This function puts a copy of c into S if one like it isn't already there 
- */
-static void insert_S(connector_table *ct, Connector * c)
-{
-	int h;
-	Connector * e;
-
-	h = hash_S(ct, c);
-
-	for (e = ct->table[h]; e != NULL; e = e->next)
-	{
-		if ((strcmp(c->string, e->string) == 0) &&
-			(c->label == e->label) && (c->priority == e->priority))
-		{
-			return;
-		}
-	}
-	e = connector_new();
-	*e = *c;
-	e->next = ct->table[h];
-	ct->table[h] = e;
-}
-
-
-static connector_table * connector_table_new(size_t sz)
-{
-	int i;
-	connector_table *ct;
-
-	ct = (connector_table *) malloc (sizeof(connector_table));
-	ct->s_table_size = sz;
-
-	ct-> table = (Connector **) xalloc(sz * sizeof (Connector *));
-	for (i = 0; i < ct->s_table_size; i++)
-	{
-		ct->table[i] = NULL;
-	}
-	return ct;
-}
-
-/** 
- * Returns TRUE if c can match anything in the set S.
- * Because of the horrible kludge, prune match is assymetric, and
- * direction is '-' if this is an l->r pass, and '+' if an r->l pass.
- */
-static int matches_S(connector_table *ct, Connector * c, int dir)
-{
-	int h;
-	Connector * e;
-
-	h = hash_S(ct, c);
-	if (dir=='-') {
-		for (e = ct->table[h]; e != NULL; e = e->next) {
-			if (prune_match(0, e, c)) return TRUE;
-		}
-	} else {
-		for (e = ct->table[h]; e != NULL; e = e->next) {
-			if (prune_match(0, c, e)) return TRUE;
-		}
-	}
-	return FALSE;
-}
-
-/**
- * This removes the disjuncts that are empty from the list corresponding
- * to word w of the sentence.
- */
-static void clean_up(Sentence sent, int w)
-{
-	Disjunct head_disjunct, *d, *d1;
-
-	d = &head_disjunct;
-
-	d->next = sent->word[w].d;
-
-	while(d->next != NULL) {
-		if ((d->next->left == NULL) && (d->next->right == NULL)) {
-			d1 = d->next;
-			d->next = d1->next;
-			xfree((char *)d1, sizeof(Disjunct));
-		} else {
-			d = d->next;
-		}
-	}
-	sent->word[w].d = head_disjunct.next;
-}
-
-/** returns the number of disjuncts in the list pointed to by d */
+/** Returns the number of disjuncts in the list pointed to by d */
 static int count_disjuncts(Disjunct * d)
 {
 	int count = 0;
-	for (; d!=NULL; d=d->next) {
+	for (; d != NULL; d = d->next)
+	{
 		count++;
 	}
 	return count;
 }
 
-static int count_disjuncts_in_sentence(Sentence sent)
+/** 
+ * This function puts connector c into the connector table
+ * if one like it isn't already there.
+ */
+static void insert_connector(connector_table *ct, Connector * c)
 {
-/* returns the total number of disjuncts in the sentence */
-	int w, count;
-	count = 0;
-	for (w=0; w<sent->length; w++) {
-		count += count_disjuncts(sent->word[w].d);
+	int h;
+	Connector * e;
+
+	h = hash_S(c);
+
+	for (e = ct[h]; e != NULL; e = e->tableNext)
+	{
+		if ((strcmp(c->string, e->string) == 0) && 
+		    (c->label == e->label) && 
+		    (c->priority == e->priority))
+			return;
 	}
-	return count;
+	c->tableNext = ct[h];
+	ct[h] = c;
 }
 
 void prune(Sentence sent)
 {
-	int N_deleted;
-	Disjunct *d;
-	Connector *e;
+	Connector *e, *f;
 	int w;
+	int N_deleted;
+	Connector *ct[CONTABSZ];
+	Disjunct fake_head, *d, *d1;
 
-	connector_table *ct;
-
-	ct = connector_table_new (next_power_of_two_up(count_disjuncts_in_sentence(sent)));
-
-/* You know, I don't think this makes much sense.  This is probably much  */
-/* too big.  There are many fewer connectors than disjuncts.			  */
-
-	N_deleted = 1;  /* a lie to make it always do at least 2 passes */
+	/* XXX why is this here ?? */
 	count_set_effective_distance(sent);
 
-	for (;;) {
-		/* left-to-right pass */
+	N_deleted = 1;  /* a lie to make it always do at least 2 passes */
+	while(1)
+	{
+		/* Left-to-right pass */
+		zero_connector_table(ct);
 
-		for (w = 0; w < sent->length; w++) {
-			for (d = sent->word[w].d; d != NULL; d = d->next) {
-				for (e = d->left; e != NULL; e = e->next) {
-					if (!matches_S(ct, e, '-')) break;
+		/* For every word */
+		for (w = 0; w < sent->length; w++)
+		{
+			d = &fake_head;
+			d->next = sent->word[w].d;
+
+			/* For every disjunct of word */
+			while ((d1 = d->next))
+			{
+				e = d1->left;
+
+				/* For every left clause of this disjunct */
+				while (e)
+				{
+					int h = hash_S(e);
+					for (f = ct[h]; f != NULL; f = f->tableNext)
+					{
+						if (prune_match(0, f, e)) break;
+					}
+					if (!f) break; /* If f null, not a single match was found */
+					e = e->next;
 				}
-				if (e != NULL) {
-					/* we know this disjunct is dead */
+
+				/* We know this disjunct is dead since no match
+				 * can be found on a required clause. */
+				if (e)
+				{
 					N_deleted ++;
-					free_connectors(d->left);
-					free_connectors(d->right);
-					d->left = d->right = NULL;
+					free_connectors(d1->left);
+					free_connectors(d1->right);
+					d->next = d1->next;
+					xfree(d1, sizeof(Disjunct));
 				}
-			}
-			clean_up(sent, w);
-			for (d = sent->word[w].d; d != NULL; d = d->next) {
-				for (e = d->right; e != NULL; e = e->next) {
-					insert_S(ct, e);
+				else
+				{
+			 		/* Store surviving disjunct in hash table */
+					for (e = d1->right; e != NULL; e = e->next)
+					{
+						insert_connector(ct, e);
+					}
+					d = d1; /* move on to next disjunct*/
 				}
 			}
+			sent->word[w].d = fake_head.next;
 		}
 
-		if (verbosity > 2) {
-			printf("l->r pass removed %d\n",N_deleted);
+		if (2 < verbosity)
+		{
+			printf("l->r pass removed %d\n", N_deleted);
 			print_disjunct_counts(sent);
 		}
-		free_S(ct);
+
+		/* We did nothing (and this is not the 1st pass) */
 		if (N_deleted == 0) break;
 
-		/* right-to-left pass */
+		/* Right-to-left pass */
+		zero_connector_table(ct);
 		N_deleted = 0;
-		for (w = sent->length-1; w >= 0; w--) {
-			for (d = sent->word[w].d; d != NULL; d = d->next) {
-				for (e = d->right; e != NULL; e = e->next) {
-					if (!matches_S(ct, e,'+')) break;
+
+		/* For every word */
+		for (w = sent->length-1; w >= 0; w--)
+		{
+			d = &fake_head;
+			d->next = sent->word[w].d;
+
+			while ((d1 = d->next))
+			{
+				e = d1->right;
+
+				while (e)
+				{
+					int h = hash_S(e);
+					for (f = ct[h]; f != NULL; f = f->tableNext)
+					{
+						if (prune_match(0, e, f)) break;
+					}
+					if (!f) break; /* If f null, not a single match was found */
+					e = e->next;
 				}
-				if (e != NULL) {
-					/* we know this disjunct is dead */
+
+				/* We know this disjunct is dead since it can't match
+				 * to the right*/
+				if(e)
+				{
 					N_deleted ++;
-					free_connectors(d->left);
-					free_connectors(d->right);
-					d->left = d->right = NULL;
+					free_connectors(d1->left);
+					free_connectors(d1->right);
+					d->next = d1->next;
+					xfree(d1, sizeof(Disjunct));
 				}
-			}
-			clean_up(sent, w);
-			for (d = sent->word[w].d; d != NULL; d = d->next) {
-				for (e = d->left; e != NULL; e = e->next) {
-					insert_S(ct, e);
+				else
+				{
+					/* Store surviving disjunct in hash table */
+					for (e = d1->left; e != NULL; e = e->next)
+					{
+						insert_connector(ct, e);
+					}
+					d = d1; /* move on to next disjunct*/
 				}
+				sent->word[w].d = fake_head.next;
 			}
 		}
-		if (verbosity > 2) {
-			printf("r->l pass removed %d\n",N_deleted);
+
+		if (verbosity > 2)
+		{
+			printf("r->l pass removed %d\n", N_deleted);
 			print_disjunct_counts(sent);
 		}
-		free_S(ct);
+
+		/* We made no change on this pass */
 		if (N_deleted == 0) break;
 		N_deleted = 0;
 	}
-	connector_table_delete(ct);
 }
 
 /*
@@ -452,7 +437,7 @@ void prune(Sentence sent)
 /**
  * hash function that takes a string and a seed value i
  */
-static int string_hash(disjunct_dup_table *dt, const char * s, int i) 
+static int string_hash(disjunct_dup_table *dt, const char * s, int i)
 {
 	for(;*s != '\0';s++) i = i + (i<<1) + randtable[(*s + i) & (RTSIZE-1)];
 	return (i & (dt->dup_table_size-1));
@@ -466,80 +451,84 @@ static int string_hash(disjunct_dup_table *dt, const char * s, int i)
   been implemented below.  There are three problems with it:
 
   (1) It is almost never the case that any disjuncts are eliminated.
-	  (The code below has works correctly with fat links, but because
-	  all of the fat connectors on a fat disjunct have the same matching
-	  string, the only time a disjuct will die is if it is the same
-	  as another one.  This is captured by the simplistic version below.
+      (The code below has works correctly with fat links, but because
+      all of the fat connectors on a fat disjunct have the same matching
+      string, the only time a disjuct will die is if it is the same
+      as another one.  This is captured by the simplistic version below.
 
   (2) connector_matches_alam may not be exactly correct.  I don't
-	  think it does the fat link matches properly.   (See the comment
-	  in and.c for more information about matching fat links.)  This is
-	  irrelevant because of (1).
-	
+      think it does the fat link matches properly.   (See the comment
+      in and.c for more information about matching fat links.)  This is
+      irrelevant because of (1).
+   
   (3) The linkage that is eliminated by this, might just be the one that
-	  passes post-processing, as the following example shows.
-	  This is pretty silly, and should probably be changed.
+      passes post-processing, as the following example shows.
+      This is pretty silly, and should probably be changed.
 
 > telling John how our program works would be stupid
 Accepted (2 linkages, 1 with no P.P. violations)
   Linkage 1, cost vector = (0, 0, 7)
 
-	+------------------G-----------------+		
-	+-----R-----+----CL----+			 |		
-	+---O---+   |   +---D--+---S---+	 +--I-+-AI-+
-	|	   |   |   |	  |	   |	 |	|	|
+   +------------------G-----------------+      
+   +-----R-----+----CL----+             |      
+   +---O---+   |   +---D--+---S---+     +--I-+-AI-+
+   |       |   |   |      |       |     |    |    |
 telling.g John how our program.n works would be stupid
 
-			   /////		  CLg	 <---CLg--->  CL		telling.g
- (g)		   telling.g	  G	   <---G----->  G		 would
- (g) (d)	   telling.g	  R	   <---R----->  R		 how
- (g) (d)	   telling.g	  O	   <---O----->  O		 John
- (g) (d)	   how			CLe	 <---CLe--->  CL		program.n
- (g) (d) (e)   our			D	   <---Ds---->  Ds		program.n
- (g) (d) (e)   program.n	  Ss	  <---Ss---->  Ss		works
- (g)		   would		  I	   <---Ix---->  Ix		be
- (g)		   be			 AI	  <---AIi--->  AIi	   stupid
+               /////         CLg    <---CLg--->  CL      telling.g
+ (g)           telling.g     G      <---G----->  G       would
+ (g) (d)       telling.g     R      <---R----->  R       how
+ (g) (d)       telling.g     O      <---O----->  O       John
+ (g) (d)       how           CLe    <---CLe--->  CL      program.n
+ (g) (d) (e)   our           D      <---Ds---->  Ds      program.n
+ (g) (d) (e)   program.n     Ss     <---Ss---->  Ss      works
+ (g)           would         I      <---Ix---->  Ix      be
+ (g)           be            AI     <---AIi--->  AIi     stupid
 
 (press return for another)
 >
   Linkage 2 (bad), cost vector = (0, 0, 7)
 
-	+------------------G-----------------+		
-	+-----R-----+----CL----+			 |		
-	+---O---+   |   +---D--+---S---+	 +--I-+-AI-+
-	|	   |   |   |	  |	   |	 |	|	|
+   +------------------G-----------------+      
+   +-----R-----+----CL----+             |      
+   +---O---+   |   +---D--+---S---+     +--I-+-AI-+
+   |       |   |   |      |       |     |    |    |
 telling.g John how our program.n works would be stupid
 
-			   /////		  CLg	 <---CLg--->  CL		telling.g
- (g)		   telling.g	  G	   <---G----->  G		 would
- (g) (d)	   telling.g	  R	   <---R----->  R		 how
- (g) (d)	   telling.g	  O	   <---O----->  O		 John
- (g) (d)	   how			CLe	 <---CLe--->  CL		program.n
- (g) (d) (e)   our			D	   <---Ds---->  Ds		program.n
- (g) (d) (e)   program.n	  Ss	  <---Ss---->  Ss		works
- (g)		   would		  I	   <---Ix---->  Ix		be
- (g)		   be			 AI	  <---AI---->  AI		stupid
+               /////         CLg    <---CLg--->  CL      telling.g
+ (g)           telling.g     G      <---G----->  G       would
+ (g) (d)       telling.g     R      <---R----->  R       how
+ (g) (d)       telling.g     O      <---O----->  O       John
+ (g) (d)       how           CLe    <---CLe--->  CL      program.n
+ (g) (d) (e)   our           D      <---Ds---->  Ds      program.n
+ (g) (d) (e)   program.n     Ss     <---Ss---->  Ss      works
+ (g)           would         I      <---Ix---->  Ix      be
+ (g)           be            AI     <---AI---->  AI      stupid
 
 P.P. violations:
-		Special subject rule violated
+      Special subject rule violated
 */
 
-int connector_matches_alam(Connector * a, Connector * b) {
-/* This returns true if the connector a matches everything that b
-   matches, and possibly more.  (alam=at least as much)
-
-   TRUE for equal connectors.
-   remains TRUE if multi-match added to the first.
-   remains TRUE if subsrcripts deleted from the first.
-
-*/
+/**
+ * This returns true if the connector a matches everything that b
+ *  matches, and possibly more.  (alam=at least as much)
+ *
+ * TRUE for equal connectors.
+ * remains TRUE if multi-match added to the first.
+ * remains TRUE if subsrcripts deleted from the first.
+ */
+int connector_matches_alam(Connector * a, Connector * b)
+{
 	char * s, * t, *u;
 	if (((!a->multi) && b->multi) ||
 		(a->label != b->label) ||
 		(a->priority != b->priority))  return FALSE;
 	s = a->string;
 	t = b->string;
-	while(isupper(*s) || isupper(*t)) {
+
+	/* isupper -- connectors cannot be UTF8 at this time */
+	while(isupper(*s) || isupper(*t))
+	{
 		if (*s == *t) {
 			s++;
 			t++;
@@ -567,7 +556,7 @@ static inline int pconnector_hash(disjunct_dup_table *dt, Connector * c, int i)
 }
 
 /**
- * This is a hash function for disjuncts 
+ * This is a hash function for disjuncts
  */
 static int hash_disjunct(disjunct_dup_table *dt, Disjunct * d)
 {
@@ -587,7 +576,7 @@ static int hash_disjunct(disjunct_dup_table *dt, Disjunct * d)
 
 /**
  * Returns TRUE if disjunct d1 can match anything that d2 can
- * if this happens, it constitutes a proof that there is absolutely 
+ * if this happens, it constitutes a proof that there is absolutely
  * no use for d2.
  */
 int disjunct_matches_alam(Disjunct * d1, Disjunct * d2)
@@ -688,7 +677,7 @@ Disjunct * eliminate_duplicate_disjuncts(Disjunct * d)
  * This is the old version that doesn't check for domination, just
  * equality.
  */
-static int old_hash_disjunct(disjunct_dup_table *dt, Disjunct * d) 
+static int old_hash_disjunct(disjunct_dup_table *dt, Disjunct * d)
 {
 	int i;
 	Connector *e;
@@ -814,48 +803,34 @@ Disjunct * eliminate_duplicate_disjuncts(Disjunct * d)
 
 
 /* ================================================================= */
-/* Here is expression pruning.  This is done even before the expressions
-   are turned into lists of disjuncts.
-
-   This uses many of the same data structures and functions that are used
-   by prune.
-*/
-
 /**
- * Computes and returns the number of connectors in all of the expressions
- * of the sentence.
+ * Here is expression pruning.  This is done even before the expressions
+ * are turned into lists of disjuncts.
+ *
+ * This uses many of the same data structures and functions that are used
+ * by prune.
+ *
+ * The purge operations remove all irrelevant stuff from the expression,
+ * and free the purged stuff.  A connector is deemed irrelevant if its
+ * string pointer has been set to NULL.  The passes through the sentence
+ * have the job of doing this.
+ *
+ * If an OR or AND type expression node has one child, we can replace it
+ * by its child.  This, of course, is not really necessary, except for
+ * performance(?)
  */
-static int size_of_sentence_expressions(Sentence sent)
-{
-	X_node * x;
-	int w, size;
-	size = 0;
-	for (w=0; w<sent->length; w++) {
-		for (x=sent->word[w].x; x!=NULL; x = x->next) {
-			size += size_of_expression(x->exp);
-		}
-	}
-	return size;
-}
-
-/* The purge operations remove all irrelevant stuff from the expression,	*/
-/* and free the purged stuff.  A connector is deemed irrelevant if its	  */
-/* string pointer has been set to NULL.  The passes through the sentence	*/
-/* have the job of doing this.											  */
-
-/* If an OR or AND type expression node has one child, we can replace it by */
-/* its child.  This, of course, is not really necessary					 */
 
 static Exp* purge_Exp(Exp *);
 
 /**
- * get rid of the elements with null expressions
+ * Get rid of the elements with null expressions
  */
 static E_list * or_purge_E_list(E_list * l)
 {
 	E_list * el;
 	if (l == NULL) return NULL;
-	if ((l->e = purge_Exp(l->e)) == NULL) {
+	if ((l->e = purge_Exp(l->e)) == NULL)
+	{
 		el = or_purge_E_list(l->next);
 		xfree((char *)l, sizeof(E_list));
 		return el;
@@ -871,12 +846,14 @@ static E_list * or_purge_E_list(E_list * l)
 static int and_purge_E_list(E_list * l)
 {
 	if (l == NULL) return 1;
-	if ((l->e = purge_Exp(l->e)) == NULL) {
+	if ((l->e = purge_Exp(l->e)) == NULL)
+	{
 		free_E_list(l->next);
 		xfree((char *)l, sizeof(E_list));
 		return 0;
 	}
-	if (and_purge_E_list(l->next) == 0) {
+	if (and_purge_E_list(l->next) == 0)
+	{
 		free_Exp(l->e);
 		xfree((char *)l, sizeof(E_list));
 		return 0;
@@ -890,22 +867,31 @@ static int and_purge_E_list(E_list * l)
  */
 static Exp* purge_Exp(Exp *e)
 {
-	if (e->type == CONNECTOR_type) {
-		if (e->u.string == NULL) {
+	if (e->type == CONNECTOR_type)
+	{
+		if (e->u.string == NULL)
+		{
 			xfree((char *)e, sizeof(Exp));
 			return NULL;
-		} else {
+		}
+		else
+		{
 			return e;
 		}
 	}
-	if (e->type == AND_type) {
-		if (and_purge_E_list(e->u.l) == 0) {
+	if (e->type == AND_type)
+	{
+		if (and_purge_E_list(e->u.l) == 0)
+		{
 			xfree((char *)e, sizeof(Exp));
 			return NULL;
 		}
-	} else {
+	}
+	else
+	{
 		e->u.l = or_purge_E_list(e->u.l);
-		if (e->u.l == NULL) {
+		if (e->u.l == NULL)
+		{
 			xfree((char *)e, sizeof(Exp));
 			return NULL;
 		}
@@ -917,7 +903,8 @@ static Exp* purge_Exp(Exp *e)
    The code is excised for these reasons.
 */
 /*
-	if ((e->u.l != NULL) && (e->u.l->next == NULL)) {
+	if ((e->u.l != NULL) && (e->u.l->next == NULL))
+	{
 		ne = e->u.l->e;
 		xfree((char *) e->u.l, sizeof(E_list));
 		xfree((char *) e, sizeof(Exp));
@@ -927,6 +914,32 @@ static Exp* purge_Exp(Exp *e)
 	return e;
 }
 
+/**
+ * Returns TRUE if c can match anything in the set S.
+ */
+static inline int matches_S(connector_table *ct, Connector * c, int dir)
+{
+	Connector * e;
+	int h = hash_S(c);
+
+	if (dir == '-')
+	{
+		for (e = ct[h]; e != NULL; e = e->tableNext)
+		{
+			if (prune_match(0, e, c)) return TRUE;
+		}
+		return FALSE;
+	}
+	else
+	{
+		for (e = ct[h]; e != NULL; e = e->tableNext)
+		{
+			if (prune_match(0, c, e)) return TRUE;
+		}
+		return FALSE;
+	}
+}
+
 /**
  * Mark as dead all of the dir-pointing connectors
  * in e that are not matched by anything in the current set.
@@ -934,50 +947,65 @@ static Exp* purge_Exp(Exp *e)
  */
 static int mark_dead_connectors(connector_table *ct, Exp * e, int dir)
 {
-	Connector dummy;
 	int count;
-	E_list *l;
-	init_connector(&dummy);
-	dummy.label = NORMAL_LABEL;
-	dummy.priority = THIN_priority;
 	count = 0;
-	if (e->type == CONNECTOR_type) {
-		if (e->dir == dir) {
+	if (e->type == CONNECTOR_type)
+	{
+		if (e->dir == dir)
+		{
+			Connector dummy;
+			init_connector(&dummy);
+			dummy.label = NORMAL_LABEL;
+			dummy.priority = THIN_priority;
 			dummy.string = e->u.string;
-			if (!matches_S(ct, &dummy,dir)) {
+			if (!matches_S(ct, &dummy, dir))
+			{
 				e->u.string = NULL;
 				count++;
 			}
 		}
-	} else {
-		for (l=e->u.l; l!=NULL; l=l->next) {
+	}
+	else
+	{
+		E_list *l;
+		for (l = e->u.l; l != NULL; l = l->next)
+		{
 			count += mark_dead_connectors(ct, l->e, dir);
 		}
 	}
 	return count;
 }
 
-/** 
+/**
  * Put into the set S all of the dir-pointing connectors still in e.
+ * Return a list of allocated dummy connectors; these will need to be
+ * freed.
  */
-static void insert_connectors(connector_table *ct, Exp * e, int dir)
+static Connector * insert_connectors(connector_table *ct, Exp * e,
+                                     Connector *alloc_list, int dir)
 {
-	Connector dummy;
-	E_list *l;
-	init_connector(&dummy);
-	dummy.label = NORMAL_LABEL;
-	dummy.priority = THIN_priority;
-
-	if (e->type == CONNECTOR_type) {
-		if (e->dir == dir) {
-			dummy.string = e->u.string;
-			insert_S(ct, &dummy);
-		}
-	} else {
-		for (l=e->u.l; l!=NULL; l=l->next) {
-			insert_connectors(ct, l->e, dir);
+	if (e->type == CONNECTOR_type)
+	{
+		if (e->dir == dir)
+		{
+			Connector *dummy = connector_new();
+			dummy->label = NORMAL_LABEL;
+			dummy->priority = THIN_priority;
+			dummy->string = e->u.string;
+			insert_connector(ct, dummy);
+			dummy->next = alloc_list;
+			alloc_list = dummy;
 		}
 	}
+	else
+	{
+		E_list *l;
+		for (l=e->u.l; l!=NULL; l=l->next)
+		{
+			alloc_list = insert_connectors(ct, l->e, alloc_list, dir);
+		}
+	}
+	return alloc_list;
 }
 
 /**
@@ -989,12 +1017,16 @@ static void clean_up_expressions(Sentence sent, int w)
 	X_node head_node, *d, *d1;
 	d = &head_node;
 	d->next = sent->word[w].x;
-	while(d->next != NULL) {
-		if (d->next->exp == NULL) {
+	while (d->next != NULL)
+	{
+		if (d->next->exp == NULL)
+		{
 			d1 = d->next;
 			d->next = d1->next;
 			xfree((char *)d1, sizeof(X_node));
-		} else {
+		}
+		else
+		{
 			d = d->next;
 		}
 	}
@@ -1006,67 +1038,90 @@ void expression_prune(Sentence sent)
 	int N_deleted;
 	X_node * x;
 	int w;
-	connector_table *ct;
+	Connector *ct[CONTABSZ];
+	Connector *dummy_list = NULL;
 
-	ct = connector_table_new(next_power_of_two_up(size_of_sentence_expressions(sent)));
+	zero_connector_table(ct);
 
 	N_deleted = 1;  /* a lie to make it always do at least 2 passes */
 
-	for (;;) {
-		/* left-to-right pass */
-		for (w = 0; w < sent->length; w++) {
-			for (x = sent->word[w].x; x != NULL; x = x->next) {
-/*	 printf("before marking: "); print_expression(x->exp); printf("\n"); */
+	while(1)
+	{
+		/* Left-to-right pass */
+		/* For every word */
+		for (w = 0; w < sent->length; w++)
+		{
+			/* For every expression in word */
+			for (x = sent->word[w].x; x != NULL; x = x->next)
+			{
+/* printf("before marking: "); print_expression(x->exp); printf("\n"); */
 				N_deleted += mark_dead_connectors(ct, x->exp, '-');
-/*	 printf("after marking marking: "); print_expression(x->exp); printf("\n"); */
+/* printf("after marking marking: "); print_expression(x->exp); printf("\n"); */
 			}
-			for (x = sent->word[w].x; x != NULL; x = x->next) {
-/*	 printf("before purging: "); print_expression(x->exp); printf("\n"); */
+			for (x = sent->word[w].x; x != NULL; x = x->next)
+			{
+/* printf("before purging: "); print_expression(x->exp); printf("\n"); */
 				x->exp = purge_Exp(x->exp);
-/*	 printf("after purging: "); print_expression(x->exp); printf("\n"); */
+/* printf("after purging: "); print_expression(x->exp); printf("\n"); */
 			}
-			clean_up_expressions(sent, w);  /* gets rid of X_nodes with NULL exp */
-			for (x = sent->word[w].x; x != NULL; x = x->next) {
-				insert_connectors(ct, x->exp,'+');
+
+			/* gets rid of X_nodes with NULL exp */
+			clean_up_expressions(sent, w);
+			for (x = sent->word[w].x; x != NULL; x = x->next)
+			{
+				dummy_list = insert_connectors(ct, x->exp, dummy_list, '+');
 			}
 		}
 
-		if (verbosity > 2) {
-			printf("l->r pass removed %d\n",N_deleted);
+		if (verbosity > 2)
+		{
+			printf("l->r pass removed %d\n", N_deleted);
 			print_expression_sizes(sent);
 		}
 
-		free_S(ct);
+		/* Free the allocated dummy connectors */
+		free_connectors(dummy_list);
+		dummy_list = NULL;
+		zero_connector_table(ct);
+
 		if (N_deleted == 0) break;
 
-		/* right-to-left pass */
+		/* Right-to-left pass */
 		N_deleted = 0;
-		for (w = sent->length-1; w >= 0; w--) {
-			for (x = sent->word[w].x; x != NULL; x = x->next) {
+		for (w = sent->length-1; w >= 0; w--)
+		{
+			for (x = sent->word[w].x; x != NULL; x = x->next)
+			{
 /*	 printf("before marking: "); print_expression(x->exp); printf("\n"); */
 				N_deleted += mark_dead_connectors(ct, x->exp, '+');
 /*	 printf("after marking: "); print_expression(x->exp); printf("\n"); */
 			}
-			for (x = sent->word[w].x; x != NULL; x = x->next) {
+			for (x = sent->word[w].x; x != NULL; x = x->next)
+			{
 /*	 printf("before perging: "); print_expression(x->exp); printf("\n"); */
 				x->exp = purge_Exp(x->exp);
 /*	 printf("after perging: "); print_expression(x->exp); printf("\n"); */
 			}
 			clean_up_expressions(sent, w);  /* gets rid of X_nodes with NULL exp */
-			for (x = sent->word[w].x; x != NULL; x = x->next) {
-				insert_connectors(ct, x->exp, '-');
+			for (x = sent->word[w].x; x != NULL; x = x->next)
+			{
+				dummy_list = insert_connectors(ct, x->exp, dummy_list, '-');
 			}
 		}
 
-		if (verbosity > 2) {
-			printf("r->l pass removed %d\n",N_deleted);
+		if (verbosity > 2)
+		{
+			printf("r->l pass removed %d\n", N_deleted);
 			print_expression_sizes(sent);
 		}
-		free_S(ct);
+
+		/* Free the allocated dummy connectors */
+		free_connectors(dummy_list);
+		dummy_list = NULL;
+		zero_connector_table(ct);
 		if (N_deleted == 0) break;
 		N_deleted = 0;
 	}
-	connector_table_delete(ct);
 }
 
 
@@ -1121,7 +1176,7 @@ void expression_prune(Sentence sent)
    deletable, this is equivalent to RUTHLESS.   --DS, 7/97
 */
 
-/** 
+/**
  * returns the number of connectors in the left lists of the disjuncts.
  */
 static int left_connector_count(Disjunct * d)
@@ -1178,7 +1233,7 @@ static void power_table_delete(power_table *pt)
 	free(pt);
 }
 
-/** 
+/**
  * This hash function only looks at the leading upper case letters of
  * the connector string, and the label fields.  This ensures that if two
  * strings match (formally), then they must hash to the same place.
@@ -1224,7 +1279,7 @@ static int set_dist_fields(Connector * c, int w, int delta)
 	return i;
 }
 
-/** 
+/**
  * Allocates and builds the initial power hash tables
  */
 static power_table * power_table_new(Sentence sent)
@@ -1465,7 +1520,7 @@ static int left_connector_list_update(prune_context *pc, Connector *c, int word_
  * If it does find a way to match it, it updates the c->word fields
  * correctly.
  */
-static int right_connector_list_update(prune_context *pc, Sentence sent, Connector *c, 
+static int right_connector_list_update(prune_context *pc, Sentence sent, Connector *c,
                                        int word_c, int w, int shallow)
 {
 	int n;
@@ -1502,7 +1557,7 @@ int power_prune(Sentence sent, int mode, Parse_Options opts)
 
 	pc = (prune_context *) malloc (sizeof(prune_context));
 	pc->power_cost = 0;
-	pc->power_prune_mode = mode; 
+	pc->power_prune_mode = mode;
 	pc->null_links = (opts->min_null_count > 0);
 	pc->N_changed = 1;  /* forces it always to make at least two passes */
 	pc->deletable = sent->deletable;
