@@ -122,14 +122,18 @@ static void dict_error2(Dictionary dict, const char * s, const char *s2)
 	}
 	if (s2)
 	{
-		prt_error("Error parsing dictionary %s.\n"
+		err_ctxt ec;
+		ec.sent = NULL;
+		err_msg(&ec, Error, "Error parsing dictionary %s.\n"
 		          "%s %s\n\t line %d, tokens = %s\n",
 		        dict->name,
 		        s, s2, dict->line_number, tokens);
 	}
 	else
 	{
-		prt_error("Error parsing dictionary %s.\n"
+		err_ctxt ec;
+		ec.sent = NULL;
+		err_msg(&ec, Error, "Error parsing dictionary %s.\n"
 		          "%s\n\t line %d, tokens = %s\n",
 		        dict->name,
 		        s, dict->line_number, tokens);
@@ -144,9 +148,11 @@ static void dict_error(Dictionary dict, const char * s)
 
 static void warning(Dictionary dict, const char * s)
 {
-	prt_error("Warning: %s\n"
-	          "\tline %d, current token = \"%s\"\n",
-	          s, dict->line_number, dict->token);
+	err_ctxt ec;
+	ec.sent = NULL;
+	err_msg(&ec, Warn, "Warning: %s\n"
+	        "\tline %d, current token = \"%s\"\n",
+	        s, dict->line_number, dict->token);
 }
 
 /**
@@ -378,7 +384,7 @@ int dict_order(char *s, char *t)
 */
 
 /* terse version */
-static int dict_order(const char *s, const char *t)
+static inline int dict_order(const char *s, const char *t)
 {
 	while (*s != '\0' && *s == *t) {s++; t++;}
 	return (((*s == '.')?(1):((*s)<<1))  -  ((*t == '.')?(1):((*t)<<1)));
@@ -396,7 +402,7 @@ static int dict_order(const char *s, const char *t)
  * by "\0", and take the difference.  This behavior matches that
  * of the function dict_order().
  */
-static int dict_order_wild(const char * s, const char * t)
+static inline int dict_order_wild(const char * s, const char * t)
 {
 	while((*s != '\0') && (*s == *t)) {s++; t++;}
 	if ((*s == '*') || (*t == '*')) return 0;
@@ -542,6 +548,9 @@ static Dict_node * rdictionary_lookup(Dict_node *llist,
  * dictionary_lookup_list() - return lookup list of words in the dictionary
  *
  * Returns a pointer to a lookup list of the words in the dictionary.
+ * Matches include word that appear in idioms.  Use
+ * abridged_lookup_list() to obtain matches, excluding idioms.
+ *
  * This list is made up of Dict_nodes, linked by their right pointers.
  * The node, file and string fields are copied from the dictionary.
  *
@@ -554,7 +563,19 @@ Dict_node * dictionary_lookup_list(Dictionary dict, const char *s)
    return llist;
 }
 
-static Dict_node * abridged_lookup_list(Dictionary dict, const char *s)
+/**
+ * abridged_lookup_list() - return lookup list of words in the dictionary
+ *
+ * Returns a pointer to a lookup list of the words in the dictionary.
+ * Excludes any idioms that contain the word; use
+ * dictionary_lookup_list() to obtain the complete list.
+ *
+ * This list is made up of Dict_nodes, linked by their right pointers.
+ * The node, file and string fields are copied from the dictionary.
+ *
+ * The returned list must be freed with free_lookup_list().
+ */
+Dict_node * abridged_lookup_list(Dictionary dict, const char *s)
 {
 	Dict_node *llist;
    llist = rdictionary_lookup(NULL, dict->root, s, FALSE);
@@ -970,6 +991,168 @@ static Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok)
 
 #endif
 
+/* ======================================================================== */
+/* Tree balancing utilities, used to implement an AVL tree.
+ * Unfortunately, AVL tree insertion is very slowww, unusably
+ * slow for creating the dictionary. The code is thus ifdef'ed out
+ * but is left here for debugging and other sundry purposes.
+ */
+
+static Dict_node *rotate_right(Dict_node *root)
+{
+	Dict_node *pivot = root->left;
+	root->left = pivot->right;
+	pivot->right = root;
+	return pivot;
+}
+
+#ifdef USE_AVL_TREE_FOR_INSERTION
+
+static Dict_node *rotate_left(Dict_node *root)
+{
+	Dict_node *pivot = root->right;
+	root->right = pivot->left;
+	pivot->left = root;
+	return pivot;
+}
+
+/* Return tree height. XXX this is not trail-recursive! */
+static int tree_depth (Dict_node *n)
+{
+	int l, r;
+	if (NULL == n) return 0;
+	if (NULL == n->left) return 1+tree_depth(n->right);
+	if (NULL == n->right) return 1+tree_depth(n->left);
+	l = tree_depth(n->left);
+	r = tree_depth(n->right);
+	if (l < r) return r+1;
+	return l+1;
+}
+
+static int tree_balance(Dict_node *n)
+{
+	int l = tree_depth(n->left);
+	int r = tree_depth(n->right);
+	return r-l;
+}
+
+/**
+ * Rebalance the dictionary tree.
+ * This recomputes the tree depth wayy too often, but so what.. this
+ * only wastes cpu time during the initial dictinary read.
+ */
+static Dict_node *rebalance(Dict_node *root)
+{
+	int bal = tree_balance(root);
+	if (2 == bal)
+	{
+		bal = tree_balance(root->right);
+		if (-1 == bal)
+		{
+			root->right = rotate_right (root->right);
+		}
+		return rotate_left(root);
+	}
+	else if (-2 == bal)
+	{
+		bal = tree_balance(root->left);
+		if (1 == bal)
+		{
+			root->left = rotate_left (root->left);
+		}
+		return rotate_right(root);
+	}
+	return root;
+}
+
+#endif /* USE_AVL_TREE_FOR_INSERTION */
+
+/* ======================================================================== */
+/* Implementation of the DSW algo for rebalancing a binary tree.
+ * The point is -- after building the dictionary tree, we rebalance it 
+ * once at the end. This is a **LOT LOT** quicker than maintaing an 
+ * AVL tree along the way (less than quarter-of-a-second vs. about
+ * a minute or more!) FWIW, the DSW tree is even more balanced than
+ * the AVL tree is (its less deep, more full).
+ *
+ * The DSW algo, with C++ code, is described in 
+ *
+ * Timothy J. Rolfe, "One-Time Binary Search Tree Balancing:
+ * The Day/Stout/Warren (DSW) Algorithm", inroads, Vol. 34, No. 4
+ * (December 2002), pp. 85-88
+ * http://penguin.ewu.edu/~trolfe/DSWpaper/
+ */
+
+static Dict_node * dsw_tree_to_vine (Dict_node *root)
+{
+	Dict_node *vine_tail, *vine_head, *rest;
+	Dict_node vh;
+
+	vine_head = &vh;
+	vine_head->left = NULL;
+	vine_head->right = root;
+	vine_tail = vine_head;
+	rest = root;
+
+	while (NULL != rest)
+	{
+		/* If no left, we are done, do the right */
+		if (NULL == rest->left)
+		{
+			vine_tail = rest;
+			rest = rest->right;
+		}
+	 	/* eliminate the left subtree */
+		else
+		{
+			rest = rotate_right(rest);
+			vine_tail->right = rest;
+		}
+	}
+
+	return vh.right;
+}
+
+static void dsw_compression (Dict_node *root, unsigned int count)
+{
+	unsigned int j;
+	for (j = 0; j < count; j++)
+	{
+		/* Compound left rotation */
+		Dict_node * pivot = root->right;
+		root->right = pivot->right;
+		root = pivot->right;
+		pivot->right = root->left;
+		root->left = pivot;
+	}
+}
+
+/* Return size of the full portion of the tree
+ * Gets the next pow(2,k)-1
+ */
+static inline unsigned int full_tree_size (unsigned int size)
+{
+	unsigned int pk = 1;
+	while (pk <= size) pk = 2*pk + 1;
+	return pk/2;
+}
+
+static Dict_node * dsw_vine_to_tree (Dict_node *root, int size)
+{
+	Dict_node vine_head;
+	unsigned int full_count = full_tree_size(size +1);
+
+	vine_head.left = NULL;
+	vine_head.right = root;
+
+	dsw_compression(&vine_head, size - full_count);
+	for (size = full_count ; size > 1 ; size /= 2)
+	{
+		dsw_compression(&vine_head, size / 2);
+	}
+	return vine_head.right;
+}
+
 /* ======================================================================== */
 /**
  * Insert the new node into the dictionary below node n.
@@ -977,24 +1160,37 @@ static Exp * restricted_expression(Dictionary dict, int and_ok, int or_ok)
  * Assumes that the "n" field of new is already set, and the left
  * and right fields of it are NULL.
  *
- * XXX The resulting tree is rather highly unbalanced. It would improve
- * performance of rdictionary_lookup() by (at least!) a factor of two 
- * if a tree balancing step was run after the dictionary creation! XXX
+ * The resulting tree is highly unbalanced. It needs to be rebalanced
+ * before used.
  */
 Dict_node * insert_dict(Dictionary dict, Dict_node * n, Dict_node * newnode)
 {
 	int comp;
 
-	if (n == NULL) return newnode;
+	if (NULL == n) return newnode;
 
 	comp = dict_order(newnode->string, n->string);
 	if (comp < 0)
 	{
+		if (NULL == n->left)
+		{
+			n->left = newnode;
+			return n;
+		} 
 		n->left = insert_dict(dict, n->left, newnode);
+		return n;
+		/* return rebalance(n); Uncomment to get an AVL tree */
 	}
 	else if (comp > 0)
 	{
+		if (NULL == n->right)
+		{
+			n->right = newnode;
+			return n;
+		}
 		n->right = insert_dict(dict, n->right, newnode);
+		return n;
+		/* return rebalance(n); Uncomment to get an AVL tree */
 	}
 	else
 	{
@@ -1003,7 +1199,6 @@ Dict_node * insert_dict(Dictionary dict, Dict_node * n, Dict_node * newnode)
 		dict_error(dict, t);
 		return NULL;
 	}
-	return n;
 }
 
 /**
@@ -1013,10 +1208,15 @@ Dict_node * insert_dict(Dictionary dict, Dict_node * n, Dict_node * newnode)
  * It inserts the list into the dictionary.
  * It does the middle one first, then the left half, then the right.
  *
- * XXX -- I think this insert middle, then left, then right, is a lame
- * attempt to hack around the fact that the binary tree is rather
- * badly unbalanced. Really, the correct thing to do would have been
- * to just-plain balance the tree. See the other XXX above.
+ * Note: I think this insert middle, then left, then right, has
+ * its origins as a lame attempt to hack around the fact that the 
+ * resulting binary tree is rather badly unbalanced. This has been 
+ * fixed by using the DSW rebalancing algo. Now, that would seem
+ * to render this crazy bisected-insertion algo obsoloete, but ..
+ * oddly enough, it seems to make the DSW balancing go really fast!
+ * Faster than a simple insertion. Go figure. I think this has
+ * something to do with the fact that the dicationaries are in
+ * alphabetical order! This subdivision helps randomize a bit.
  */
 static void insert_list(Dictionary dict, Dict_node * p, int l)
 {
@@ -1042,7 +1242,9 @@ static void insert_list(Dictionary dict, Dict_node * p, int l)
 	}
 	else if (is_idiom_word(dn->string))
 	{
-		prt_error("Warning: Word \"%s\" found near line %d.\n"
+		err_ctxt ec;
+		ec.sent = NULL;
+		err_msg(&ec, Warn, "Warning: Word \"%s\" found near line %d.\n"
 		        "\tWords ending \".Ix\" (x a number) are reserved for idioms.\n"
 		        "\tThis word will be ignored.\n",
 		        dn->string, dict->line_number);
@@ -1051,7 +1253,9 @@ static void insert_list(Dictionary dict, Dict_node * p, int l)
 	else if ((dn_head = abridged_lookup_list(dict, dn->string)) != NULL)
 	{
 		Dict_node *dnx;
-		prt_error("Warning: The word \"%s\" "
+		err_ctxt ec;
+		ec.sent = NULL;
+		err_msg(&ec, Warn, "Warning: The word \"%s\" "
 		          "found near line %d of %s matches the following words:\n",
 	             dn->string, dict->line_number, dict->name);
 		for (dnx = dn_head; dnx != NULL; dnx = dnx->right) {
@@ -1097,7 +1301,9 @@ static int read_entry(Dictionary dict)
 			dn = read_word_file(dict, dn, dict->token);
 			if (dn == NULL)
 			{
-				prt_error("Error opening word file %s\n", dict->token);
+				err_ctxt ec;
+				ec.sent = NULL;
+				err_msg(&ec, Error, "Error opening word file %s\n", dict->token);
 				return 0;
 			}
 		}
@@ -1318,6 +1524,8 @@ int read_dictionary(Dictionary dict)
 			return 0;
 		}
 	}
+	dict->root = dsw_tree_to_vine(dict->root);
+	dict->root = dsw_vine_to_tree(dict->root, dict->num_entries);
 	return 1;
 }
 
